
class ImageMetaData:
    def __init__(self, attributes, metadata):
        self.attributes = attributes
        self.metadata = metadata

    def identify_unique_image_locations(self, force_linux_path=False):
        ids_per_location = {}
        classifier_per_location = {}
        locations_per_classifier = {}

        sub_attributes = []
        metadata_per_location = {}

        idx_instance_id = self.attributes["instance_id"]
        idx_img_location = self.attributes["img_location"]
        if "classifier" in self.attributes:
            idx_classifier = self.attributes["classifier"]
        else:
            idx_classifier = None

        for attribute in self.attributes:
            if not attribute in ["instance_id", "img_location"]:
                sub_attributes.append(attribute)

        for record in self.metadata:
            instance_id = record[idx_instance_id]
            img_location = record[idx_img_location]

            if force_linux_path and "\\" in img_location:
                img_location = img_location.replace("\\", "/")

            # use first classifier associated with given image ...
            if not img_location in classifier_per_location:
                if idx_classifier is not None:
                    # use specified classifier ...
                    classifier_name = record[idx_classifier]
                else:
                    classifier_name = None
                classifier_per_location[img_location] = classifier_name

                if classifier_name not in locations_per_classifier:
                    locations_per_classifier[classifier_name] = []

                locations_per_classifier[classifier_name].append(img_location)

            # get metadata ...
            instance_metadata = [record[self.attributes[attribute]] for attribute in sub_attributes]

            if img_location in ids_per_location:
                ids_per_location[img_location].append(instance_id)
                metadata_per_location[img_location].append(instance_metadata)
            else:
                ids_per_location[img_location] = [instance_id]
                metadata_per_location[img_location] = [instance_metadata]

        return locations_per_classifier, ids_per_location, sub_attributes, metadata_per_location

    @staticmethod
    def load_metadata_file(tanv_config):
        # Load image index ...
        # (image_id, image_location, [classifier name])
        img_metadata_file = tanv_config.get("BINARY_IMAGES_METADATA")

        with open(img_metadata_file, "r", encoding="utf8") as metadata_file:
            all_metadata_lines = metadata_file.readlines()

        # pre-filter empty lines or comment lines ...
        raw_attributes = None
        filtered_metadata_lines = []
        for line in all_metadata_lines:
            line = line.strip()
            if len(line) == 0 or line[0] == "#":
                continue

            line_parts = line.split("\t")

            if raw_attributes is None:
                # first valid line should contain header ...
                raw_attributes = line_parts
            else:
                filtered_metadata_lines.append(line_parts)

        # there should be at least two lines: header + one record
        if raw_attributes is None:
            raise Exception("Image Meta-data should contain header")

        if len(filtered_metadata_lines) < 1:
            raise Exception("Image Meta-data should contain at least one record")

        metadata_attributes = {}
        for idx, value in enumerate(raw_attributes):
            metadata_attributes[value] = idx

        # attributes should contain: instance_id and img_location
        if (not "instance_id" in metadata_attributes) or (not "img_location" in metadata_attributes):
            raise Exception("Meta-data missing mandatory attributes: instance_id and/or img_location")

        # finally, validate that all meta-data lines have the same number of attributes ...
        for line in filtered_metadata_lines:
            if len(line) != len(metadata_attributes):
                raise Exception("Expecting attributes {0:s}, found values {1:s}".format(str(metadata_attributes), str(line)))

        return ImageMetaData(metadata_attributes, filtered_metadata_lines)

