
import cv2
import time
import numpy as np

from scipy.spatial.distance import cdist

from TangentV.indexing.index_parameters import IndexParameters
from TangentV.indexing.graph_index import GraphIndex
from TangentV.retrieval.graph_match_group import GraphMatchGroup
from TangentV.retrieval.graph_match import GraphMatch


class MatchHelper:
    @staticmethod
    def separate_matches_per_sequence(image_index, img_matches, group_att_idx, sequence_att_idx):
        matches_per_value = {}
        for image_id in img_matches:
            # only use the first location of the image for sequence based grouping ...
            metadata = image_index.images[image_id].source_metadata[0]

            group_value = metadata[group_att_idx]
            sequence_value = int(metadata[sequence_att_idx])

            if not group_value in matches_per_value:
                matches_per_value[group_value] = {}

            matches_per_value[group_value][sequence_value] = (image_id, img_matches[image_id])

        return matches_per_value

    @staticmethod
    def separate_matches_per_attributes(image_index, img_matches, group_atts_indices):
        matches_per_combination = {}
        for image_id in img_matches:
            # only use the first location of the image for attribute-based grouping ...
            metadata = image_index.images[image_id].source_metadata[0]

            key_values = []
            for group_att_idx in group_atts_indices:
                group_value = metadata[group_att_idx]
                key_values.append(str(group_value))

            image_key = "|".join(key_values)

            if not image_key in matches_per_combination:
                matches_per_combination[image_key] = []

            matches_per_combination[image_key].append((image_id, img_matches[image_id]))

        return matches_per_combination

    @staticmethod
    def find_compatible_pairs(match_index, candidate_graph):
        compatible_pairs = []
        per_candidate_node_index = {idx: [] for idx in range(len(candidate_graph.vertices))}
        for match_id in match_index:
            current_match = match_index[match_id]
            pending_test_ids = []
            for c_node_idx in current_match.candidate_nodes:
                # add any existing match from that candidate node index ...
                pending_test_ids += per_candidate_node_index[c_node_idx]

                # then, add itself to the index at candidate node ...
                per_candidate_node_index[c_node_idx].append(match_id)

            # separate by unique test nodes
            pending_test_ids = set(pending_test_ids)

            # test for compatible matches ....
            for other_id in pending_test_ids:
                other_match = match_index[other_id]

                compatible, total_score = current_match.check_compatibility(other_match)

                if compatible:
                    # merged_alignments = current_match.node_alignments.union(other_match.node_alignments)
                    # spatial_cost = MatchHelper.GraphsSpatialAlignmentCost(query_graph, candidate_graph, merged_alignments)
                    compatible_pairs.append((total_score, current_match, other_match))

        # sort by decreasing score ...
        compatible_pairs = sorted(compatible_pairs, key=lambda x: x[0], reverse=True)

        return compatible_pairs, per_candidate_node_index

    @staticmethod
    def merge_compatible_pairs(query_graph, candidate_graph, match_index, debug_region):
        debug_output_count = 0
        if debug_region is not None:
            d_min_x, d_max_x, d_min_y, d_max_y = debug_region
            debug_nodes = candidate_graph.selectVerticesByRegion(d_min_x, d_max_x, d_min_y, d_max_y)
        else:
            debug_nodes = []

        compatible_pairs, per_candidate_node_index = MatchHelper.find_compatible_pairs(match_index, candidate_graph)

        tempo_total_added = 0
        while len(compatible_pairs) > 0:
            total_score, match_1, match_2 = compatible_pairs[0]
            del compatible_pairs[0]

            str_id_1 = match_1.strID()
            str_id_2 = match_2.strID()

            # if matches still exist and have not yet been merged with other elements
            if str_id_1 in match_index and str_id_2 in match_index:
                # remove from index of matches
                del match_index[str_id_1]
                del match_index[str_id_2]

                # create merged match
                larger_match = GraphMatch.Merge(match_1, match_2)

                # comment later for better performance ...
                if debug_region is not None:
                    if len(set(larger_match.candidate_nodes).intersection(debug_nodes)) > 0:
                        debug_output_count += 1

                        query_img, cand_img = larger_match.getImage()
                        cand_img[:query_img.shape[0] + 4, :query_img.shape[1] + 4, :] = 0
                        cand_img[2:query_img.shape[0] + 2, 2:query_img.shape[1] + 2, :] = query_img

                        spatial_cost = MatchHelper.GraphsSpatialAlignmentCost(query_graph, candidate_graph,
                                                                              larger_match.node_alignments)

                        tempo_filename_template = "DELETE_match_grow_{0:d}_SCR_{1:.4f}_SPT_{2:.4f}.png"
                        debug_filename = tempo_filename_template.format(debug_output_count, larger_match.total_score,
                                                                        spatial_cost)
                        cv2.imwrite(debug_filename, cand_img)

                        print("\n")
                        print("To merge: ")
                        print(match_1)
                        print("with: ")
                        print(match_2)
                        print("---> OUTPUT (" + str(debug_output_count) + "):")
                        print(larger_match)

                # check compatibility and merge scores with other existing matches ...
                total_added = 0
                # ... remove current merged elements from candidate node index ...
                for c_node_idx in match_1.candidate_nodes:
                    per_candidate_node_index[c_node_idx].remove(str_id_1)
                for c_node_idx in match_2.candidate_nodes:
                    per_candidate_node_index[c_node_idx].remove(str_id_2)

                pending_test_ids = []
                larger_match_id = larger_match.strID()
                # ... now, compute potential merges for larger match ...
                for c_node_idx in larger_match.candidate_nodes:
                    # add any existing match from that candidate node index ...
                    pending_test_ids += per_candidate_node_index[c_node_idx]

                    # then, add itself to the index at candidate node ...
                    per_candidate_node_index[c_node_idx].append(larger_match_id)

                # separate by unique test nodes
                pending_test_ids = set(pending_test_ids)
                # test for compatible matches ....
                for other_id in pending_test_ids:
                    other_match = match_index[other_id]

                    compatible, total_score = larger_match.check_compatibility(other_match)

                    if compatible:
                        # merged_alignments = larger_match.node_alignments.union(other_match.node_alignments)
                        # spatial_cost = MatchHelper.GraphsSpatialAlignmentCost(query_graph, candidate_graph, merged_alignments)

                        compatible_pairs.append((total_score, larger_match, other_match))
                        total_added += 1

                # sorted compatible matches after possible additions ...
                tempo_total_added += total_added
                if total_added > 0:
                    compatible_pairs = sorted(compatible_pairs, key=lambda x: x[0], reverse=True)
                    # compatible_pairs = sorted(compatible_pairs, key=lambda x:x[0], reverse=False)

                # finally, add to index of available matches ...

                match_index[larger_match.strID()] = larger_match

        initial_matches = sorted(list(match_index.values()), key=lambda x: x.total_score, reverse=True)

        return initial_matches

    @staticmethod
    def connect_partially_compatible(query_graph, candidate_graph, initial_matches):
        merged_matches = []
        pending_matches = list(initial_matches)
        while len(pending_matches) > 0:
            current_match = pending_matches[0]
            del pending_matches[0]

            pos = 0
            while pos < len(pending_matches):
                other_match = pending_matches[pos]

                # check if could be merged ....
                compatible_alignments = current_match.check_partial_compatibility(other_match)

                if IndexParameters.Match_MaxSpatialScore is None:
                    spatially_accepted = True
                else:
                    spatial_cost = MatchHelper.GraphSpatialAlignmentRelativeCost(query_graph, candidate_graph,
                                                                                 current_match.node_alignments,
                                                                                 other_match.node_alignments)

                    spatially_accepted = spatial_cost < IndexParameters.Match_MaxSpatialScore

                if len(compatible_alignments) > 0 and spatially_accepted:
                    start_time = time.time()
                    # other match has nodes that are not in current largest match ...
                    path = GraphMatch.FindPathConnectingMatches(query_graph, candidate_graph,
                                                                current_match.node_alignments, compatible_alignments,
                                                                IndexParameters.IndexEntry_MinCosineSimilarity,
                                                                IndexParameters.Match_MaxPathLength)

                    end_time = time.time()

                    total_time = end_time - start_time
                    if len(path) > 0:
                        # create a copy of base match ...
                        base_extended = GraphMatch.Copy(current_match)
                        # add the path to the copy ....
                        base_extended.AddPath(path, [0.0] * (len(path) - 1))
                        # find which alignments from the other match are still compatible after
                        path_compatible = base_extended.check_partial_compatibility(other_match)

                        # first, extract sub-match from the other match that is compatible with current match
                        sub_match = GraphMatch.SubMatch(other_match, path_compatible)

                        # merge matches
                        new_extended = GraphMatch.Merge(base_extended, sub_match)

                        if len(path_compatible) == len(compatible_alignments):
                            # no incompatible nodes removed, submatch remains the same as original,
                            # simply use the newly merged match ...
                            current_match = new_extended
                        else:
                            # incompatible alignments found, and nodes from second match were removed ....
                            # test for a second merge in the opposite direction
                            second_compatible = other_match.check_partial_compatibility(current_match)

                            second_path = GraphMatch.FindPathConnectingMatches(
                                query_graph, candidate_graph, other_match.node_alignments, second_compatible,
                                IndexParameters.IndexEntry_MinCosineSimilarity, IndexParameters.Match_MaxPathLength)

                            second_extended = GraphMatch.Copy(other_match)
                            second_extended.AddPath(second_path, [0.0] * (len(second_path) - 1))
                            second_path_compatible = second_extended.check_partial_compatibility(current_match)

                            sub_match = GraphMatch.SubMatch(current_match, second_path_compatible)
                            second_new_extended = GraphMatch.Merge(second_extended, sub_match)

                            if new_extended.total_score >= second_new_extended.total_score:
                                # keep the first version of the match
                                current_match = new_extended
                            else:
                                # use the second version of the match ...
                                current_match = second_new_extended

                        # remove other match from list of pending
                        del pending_matches[pos]
                    else:
                        # cannot merge
                        pos += 1

                else:
                    if IndexParameters.Match_EarlyDeletion:
                        # will not be merged ... check if they overlap in candidate matches ...
                        max_other_score = other_match.total_score + (0 if pos == 0 else pending_matches[0].total_score)
                        if (current_match.candidate_nodes.intersection(other_match.candidate_nodes) and
                                (max_other_score < current_match.total_score)):

                            # even if merged with the second largest match, the final score would be smaller than
                            # current match. Greedily assume it will not be able to outgrow the current largest match
                            # that is associated to some of the same candidate nodes, and remove from the list of
                            # matches
                            del pending_matches[pos]
                        else:
                            pos += 1
                    else:
                        # simply move to next match
                        pos += 1

            merged_matches.append(current_match)

        # sort merged matches ...
        merged_matches = sorted(merged_matches, key=lambda x: x.total_score, reverse=True)

        return merged_matches

    @staticmethod
    def filter_low_score_incompatible(current_matches):
        # Filter matches that overlap better matches
        matched_candidate_nodes = set()

        accepted_matches = []
        for match in current_matches:
            if not matched_candidate_nodes.intersection(match.candidate_nodes):
                # unique candidate nodes matched, accept ...
                accepted_matches.append(match)
                matched_candidate_nodes = matched_candidate_nodes.union(match.candidate_nodes)

        return accepted_matches

    @staticmethod
    def join_disconnected_compatible(query_graph, candidate_graph, current_matches):
        joined_matches = []
        pending_matches = list(current_matches)
        while len(pending_matches) > 0:
            current_match = pending_matches[0]
            del pending_matches[0]

            pos = 0
            while pos < len(pending_matches):
                other_match = pending_matches[pos]

                # check if could be merged ....
                compatible_alignments, _ = current_match.check_disjoint_alignments(other_match)

                if not compatible_alignments:
                    # move to the next ...
                    pos += 1
                    continue

                # check spatial score
                if IndexParameters.Match_MaxDisconnectedSpatialScore is None:
                    spatially_accepted = True
                else:
                    spatial_cost = MatchHelper.GraphSpatialAlignmentRelativeCost(query_graph, candidate_graph,
                                                                                 current_match.node_alignments,
                                                                                 other_match.node_alignments)

                    spatially_accepted = spatial_cost < IndexParameters.Match_MaxDisconnectedSpatialScore

                if spatially_accepted:
                    # merge matches
                    current_match = GraphMatch.Merge(current_match, other_match)

                    # remove other match from list of pending
                    del pending_matches[pos]
                else:
                    # simply move to next match
                    pos += 1

            # print("\n")
            # print(current_match)
            # current_match.print_rec_confidences()

            joined_matches.append(current_match)

        # sort merged matches ...
        joined_matches = sorted(joined_matches, key=lambda x: x.total_score, reverse=True)

        return joined_matches

    @staticmethod
    def align_img_matches(query_graph, candidate_graph, img_matching_pairs, debug_region=None):
        np.seterr(all='raise')
        # initialize an index of matches ...
        match_index = {match.strID(): match for match in img_matching_pairs}
        img_stats = {}

        # Grow matches by greedily merging highly confident pairs that are directly connected to each other.
        start_grow_merge = time.time()
        initial_matches = MatchHelper.merge_compatible_pairs(query_graph, candidate_graph, match_index, debug_region)
        end_grow_merge = time.time()
        img_stats["time_layer2_grow_merge"] = end_grow_merge - start_grow_merge

        # try expanding matches by merging disconnected but compatible matches ...
        start_grow_connect = time.time()
        merged_matches = MatchHelper.connect_partially_compatible(query_graph, candidate_graph, initial_matches)
        end_grow_connect = time.time()
        img_stats["time_layer2_grow_connect"] = end_grow_connect - start_grow_connect
        # print(GraphMatch.DebugCounter)

        # Filter low-scoring matches that overlap better matches
        start_match_filter = time.time()
        accepted_matches = MatchHelper.filter_low_score_incompatible(merged_matches)
        end_match_filter = time.time()
        img_stats["time_layer2_match_filter"] = end_match_filter - start_match_filter

        # optional step ... Join disconnected matches that are spatially consistent
        start_join_matches = time.time()
        if IndexParameters.Layer2_JoinDisconnectedMatches:
            joined_matches = MatchHelper.join_disconnected_compatible(query_graph, candidate_graph, accepted_matches)
        else:
            # by-pass, do not join disconnected matches ...
            joined_matches = accepted_matches

        end_join_matches = time.time()
        img_stats["time_layer2_join_disconnected"] = end_join_matches - start_join_matches

        img_stats["count_layer2_initial_matches"] = len(img_matching_pairs)
        img_stats["count_layer2_grown_matches"] = len(initial_matches)
        img_stats["count_layer2_connected_matches"] = len(merged_matches)
        img_stats["count_layer2_final_matches"] = len(accepted_matches)
        img_stats["count_layer2_join_disconnected"] = len(joined_matches)

        return joined_matches, img_stats

    @staticmethod
    def group_matches_by_sequence(image_index, all_img_matches, group_by_attribute, sequence_by_attribute):
        group_att_idx = image_index.metadata_attributes.index(group_by_attribute)
        sequence_att_idx = image_index.metadata_attributes.index(sequence_by_attribute)
        matches_per_att_val = MatchHelper.separate_matches_per_sequence(image_index, all_img_matches, group_att_idx,
                                                                        sequence_att_idx)
        # check potential merges ...
        final_grouped_matches = []
        total_raw_found = 0
        total_merged_found = 0
        # .... for each group (unique value on group by attribute)
        for group_id in matches_per_att_val:
            all_group_matches = matches_per_att_val[group_id]
            group_matches = []
            active_matches = []

            # print("------------")
            # ... for each image in the sequence ... sorted by sequence att value ...
            total_raw_matches_group = 0
            for seq_idx, sequence_val in enumerate(sorted(all_group_matches.keys())):
                # image id on index, and set of current matches for this image ...
                img_id, image_matches = all_group_matches[sequence_val]

                total_raw_matches_group += len(image_matches)

                if seq_idx == 0:
                    # initialize ...
                    for match_idx, match in enumerate(image_matches):
                        match_region = match.candidate_region()

                        group = GraphMatchGroup.FromGroup({img_id: match}, {img_id: match_region}, group_id)
                        group_matches.append(group)
                        active_matches.append(match_idx)
                else:
                    # find potential overlaps ...
                    still_active = [False for idx in active_matches]

                    new_matches = []
                    to_add = []
                    for match_idx, match in enumerate(image_matches):
                        match_region = match.candidate_region()

                        # compare against active ...
                        merged = False
                        for idx, active_idx in enumerate(active_matches):
                            active_region = group_matches[active_idx].last_added_region()

                            _, active_prc, _ = GraphMatch.region_overlap(active_region, match_region)

                            # if there is enough overlap between active region and new match ...
                            if active_prc > IndexParameters.Match_MinGroupAreaOverlap:
                                to_add.append((active_idx, match_region, match))
                                still_active[idx] = True
                                merged = True
                                break

                        # if not merged with active, then it will be added as new
                        if not merged:
                            new_matches.append((match_region, match))

                    # apply added elements
                    for active_idx, match_region, match in to_add:
                        group_matches[active_idx].add_match(img_id, match, match_region)

                    # remove not active
                    # ... since elements will be removed ... we remove them from last to first
                    # ... to ensure original indices will be valid after removing each element
                    for idx in range(len(still_active) - 1, -1, -1):
                        if not still_active[idx]:
                            del active_matches[idx]

                    # add new matches ...
                    for match_region, match in new_matches:
                        next_idx = len(group_matches)

                        group = GraphMatchGroup.FromGroup({img_id: match}, {img_id: match_region}, group_id)
                        group_matches.append(group)
                        active_matches.append(next_idx)

            scored_groups = []
            for match_group in group_matches:
                keyframe_idx, max_score = match_group.max_score()
                scored_groups.append((max_score, keyframe_idx, match_group))

            final_grouped_matches += scored_groups

            total_raw_found += total_raw_matches_group
            total_merged_found += len(group_matches)

            # print("Lecture: " + lecture_id)
            # print("-> Initial Raw: " + str(total_raw_matches_lecture))
            # print("-> Final Grouped: " + str(len(lecture_matches)))
                

        # print("Global: ")
        # print("-> Initial Raw: " + str(total_raw_found))
        # print("-> Final Grouped: " + str(total_merged_found))

        query_stats = {
            "count_layer3_initial_matches": total_raw_found,
            "count_layer3_final_matches": total_merged_found,
        }

        final_grouped_matches = sorted(final_grouped_matches, key=lambda x:x[0], reverse=True)
        
        return final_grouped_matches, query_stats

    @staticmethod
    def group_matches_by_attributes(image_index, all_img_matches, group_by_attributes):
        assert isinstance(image_index, GraphIndex)

        group_indices = []
        for group_by_attribute in group_by_attributes:
            group_att_idx = image_index.metadata_attributes.index(group_by_attribute)
            group_indices.append(group_att_idx)

        # group matches by unique group attribute combinations
        matches_per_combination = MatchHelper.separate_matches_per_attributes(image_index, all_img_matches,
                                                                              group_indices)

        scored_groups = []
        total_raw_found = 0
        total_merged_found = len(matches_per_combination)
        for group_key in matches_per_combination:
            current_group = matches_per_combination[group_key]

            group_matches = {}
            group_regions = {}
            for image_id, image_matches in current_group:
                total_raw_found += len(image_matches)

                # chose only the best match from each image in the group ....
                top_index = np.argmax([match.total_score for match in image_matches])
                top_match = image_matches[top_index]

                group_matches[image_id] = top_match
                group_regions[image_id] = top_match.candidate_region()

            match_group = GraphMatchGroup.FromGroup(group_matches, group_regions, group_key)
            max_image_id, max_score = match_group.max_score()
            scored_groups.append((max_score, max_image_id, match_group))

        query_stats = {
            "count_layer3_initial_matches": total_raw_found,
            "count_layer3_final_matches": total_merged_found,
        }

        final_grouped_matches = sorted(scored_groups, key=lambda x: x[0], reverse=True)

        return final_grouped_matches, query_stats

    @staticmethod
    def group_matches_by_image(image_index, img_matches):
        assert isinstance(image_index, GraphIndex)

        total_raw_found = 0
        final_grouped_matches = []
        for img_idx in img_matches:
            total_raw_found += len(img_matches[img_idx])

            scored = [(match.total_score, match) for match in img_matches[img_idx]]
            scored = sorted(scored, key=lambda x: x[0], reverse=True)

            # only use the best match ...
            _, top_match = scored[0]

            # use top match only
            group_matches = {img_idx: top_match}
            group_regions = {img_idx: top_match.candidate_region()}
            group = GraphMatchGroup.FromGroup(group_matches, group_regions)

            final_grouped_matches.append((top_match.total_score, img_idx, group))

        query_stats = {
            "count_layer3_initial_matches": total_raw_found,
            "count_layer3_final_matches": len(final_grouped_matches),
        }

        final_grouped_matches = sorted(final_grouped_matches, key=lambda x: x[0], reverse=True)

        return final_grouped_matches, query_stats

    @staticmethod
    def group_matches_by_none(image_index, img_matches):
        assert isinstance(image_index, GraphIndex)

        total_raw_found = 0
        final_grouped_matches = []
        for img_idx in img_matches:
            total_raw_found += len(img_matches[img_idx])
            for match in img_matches[img_idx]:
                group_matches = {img_idx: match}
                group_regions = {img_idx: match.candidate_region()}
                group = GraphMatchGroup.FromGroup(group_matches, group_regions)

                final_grouped_matches.append((match.total_score, img_idx, group))

        query_stats = {
            "count_layer3_initial_matches": total_raw_found,
            "count_layer3_final_matches": total_raw_found,
        }

        final_grouped_matches = sorted(final_grouped_matches, key=lambda x: x[0], reverse=True)

        return final_grouped_matches, query_stats

    @staticmethod
    def add_precision_metric(scored_matches, precision_function):
        precision_grouped_matches = []
        for best_score, best_score_img, match_group in scored_matches:
            best_match = match_group.matches_per_image[best_score_img]

            # get precision term ...
            precision = precision_function(best_match)

            # add precision term
            precision_grouped_matches.append(([best_score, precision], best_score_img, match_group))

        # sort matches (again), now using the score vector (lexicographic-order)
        precision_grouped_matches = sorted(precision_grouped_matches, key=lambda x: x[0], reverse=True)

        return precision_grouped_matches

    @staticmethod
    def GraphsSpatialAlignmentCost(graph_1, graph_2, aligned_nodes, debug=False):
        # this is the cost of spatially aligning two graph based on a set of "anchor" points
        n_alignments = len(aligned_nodes)
        # get the vertices locations ....
        positions_g1, positions_g2 = MatchHelper.GraphAlignmentsToPositionArrays(graph_1, graph_2, aligned_nodes)

        # get best translation ...
        # use graph nodes to select a translation center ....
        # based on the node that is closest (in average) to the other nodes)
        g1_dist_matrix = cdist(positions_g1, positions_g1)
        g1_avg_distances = g1_dist_matrix.sum(axis=1) / (n_alignments - 1)
        ref_node = np.argmin(g1_avg_distances)

        # compute scale ....
        g2_dist_matrix = cdist(positions_g2, positions_g2)
        g2_avg_distances = g2_dist_matrix.sum(axis=1) / (n_alignments - 1)

        scale_factor = g1_avg_distances[ref_node] / g2_avg_distances[ref_node]

        # now center the arrays using the reference node ....
        g1_translation = np.tile(positions_g1[ref_node], (n_alignments, 1))
        g2_translation = np.tile(positions_g2[ref_node], (n_alignments, 1))

        # compute distances ....
        final_pos_g1 = positions_g1 - g1_translation
        final_pos_g2 = (positions_g2 - g2_translation) * scale_factor
        final_distances = np.sqrt(np.power(final_pos_g1 - final_pos_g2, 2).sum(axis=1))

        # use the average distance between aligned points ....
        avg_distance = final_distances.sum() / n_alignments

        # normalize ....
        avg_distance /= g1_avg_distances[ref_node]

        # comment for performance ...
        if debug:
            print("Raw Points")
            print(positions_g1)
            print(positions_g2)
            print("Raw Distance matrices:")
            print(g1_dist_matrix)
            print(g2_dist_matrix)
            print("Average Distances: ")
            print(g1_avg_distances)
            print(g2_avg_distances)
            print("Selected Ref Node: " + str(ref_node))
            print("Scale Factor: {0:.6f}".format(scale_factor))
            print("Translated/Scaled Points")
            print(final_pos_g1)
            print(final_pos_g2)
            print("Final pair-wise distances")
            print(final_distances)
            print("Final Average Distance: {0:.6f}".format(avg_distance))

        return avg_distance

    @staticmethod
    def GraphSpatialAlignmentRelativeCost(graph_1, graph_2, main_alignment, other_alignment, min_avg_dist=5.0, debug=False):
        # this is the cost of adding new alignments to the current spatial alignment ...
        positions_g1, positions_g2 = MatchHelper.GraphAlignmentsToPositionArrays(graph_1, graph_2, main_alignment)

        # get best translation ...
        # use graph nodes to select a translation center ....
        # based on the node that is closest (in average) to the other nodes)
        g1_dist_matrix = cdist(positions_g1, positions_g1)
        if len(main_alignment) > 1:
            g1_avg_distances = g1_dist_matrix.sum(axis=1) / (len(main_alignment) - 1)
        else:
            g1_avg_distances = g1_dist_matrix.sum(axis=1)
        ref_node = np.argmin(g1_avg_distances)

        # compute scale ....
        g2_dist_matrix = cdist(positions_g2, positions_g2)
        if len(main_alignment) > 1:
            g2_avg_distances = g2_dist_matrix.sum(axis=1) / (len(main_alignment) - 1)
        else:
            g2_avg_distances = g2_dist_matrix.sum(axis=1)

        if g1_avg_distances[ref_node] > min_avg_dist and g2_avg_distances[ref_node] > min_avg_dist:
            scale_factor = g1_avg_distances[ref_node] / g2_avg_distances[ref_node]
        else:
            # one average distance is too small, assume scale factor of 1.0
            scale_factor = 1.0

        # ... now apply scale and translation to the second set of alignments ...
        other_pos_g1, other_pos_g2 = MatchHelper.GraphAlignmentsToPositionArrays(graph_1, graph_2, other_alignment)

        # now center the arrays using the reference node ....
        g1_translation = np.tile(positions_g1[ref_node], (len(other_alignment), 1))
        g2_translation = np.tile(positions_g2[ref_node], (len(other_alignment), 1))

        final_pos_g1 = other_pos_g1 - g1_translation
        final_pos_g2 = (other_pos_g2 - g2_translation) * scale_factor

        final_distances = np.sqrt(np.power(final_pos_g1 - final_pos_g2, 2).sum(axis=1))

        # use the average distance between aligned points ....
        avg_distance = final_distances.sum() / len(other_alignment)

        # normalize ....
        # note that if the average distance on the first alignment is 0,
        # then the avg distance will be larger (higher spatial distortion cost)
        if g1_avg_distances[ref_node] > 0.0:
            avg_distance /= g1_avg_distances[ref_node]

        return avg_distance

    @staticmethod
    def GraphAlignmentsToPositionArrays(graph_1, graph_2, alignments):
        positions_g1 = np.zeros((len(alignments), 2), dtype=np.float64)
        positions_g2 = np.zeros((len(alignments), 2), dtype=np.float64)

        for idx, (g1_n_idx, g2_n_idx) in enumerate(alignments):
            positions_g1[idx, :] = graph_1.vertices[g1_n_idx].getCenter()
            positions_g2[idx, :] = graph_2.vertices[g2_n_idx].getCenter()

        return positions_g1, positions_g2

    @staticmethod
    def GraphAlignmentsToDiagonalLengthArrays(graph_1, graph_2, alignments):
        sizes_g1 = np.zeros((len(alignments), 2), dtype=np.float64)
        sizes_g2 = np.zeros((len(alignments), 2), dtype=np.float64)

        for idx, (g1_n_idx, g2_n_idx) in enumerate(alignments):
            sizes_g1[idx, :] = graph_1.vertices[g1_n_idx].getWidth(), graph_1.vertices[g1_n_idx].getHeight()
            sizes_g2[idx, :] = graph_2.vertices[g2_n_idx].getWidth(), graph_2.vertices[g2_n_idx].getHeight()

        diagionals_g1 = np.sqrt(np.power(sizes_g1, 2).sum(axis=1))
        diagionals_g2 = np.sqrt(np.power(sizes_g2, 2).sum(axis=1))

        return diagionals_g1, diagionals_g2

