
# =====================================================================
# Created by: Kenny Davila (February, 2018)
#
# Another interface between Tangent-V and Recognizer.
# Many of these functions should be provided by future versions of the
# recognizer itself
# =====================================================================

import math
import cv2
import numpy as np

from recognizer.symbols.math_symbol import MathSymbol
from recognizer.symbols.math_symbol_trace import MathSymbolTrace

class HandwritingRenderer:
    def __init__(self, resolution, prc_width, decay_width, decay_length, max_scaled_decay):
        self.resolution = resolution
        self.prc_width = prc_width
        self.decay_width = decay_width
        self.decay_length = decay_length
        self.max_scaled_decay = max_scaled_decay

    def render_hw_symbol(self, symbol):
        assert isinstance(symbol, MathSymbol)

        # assume the symbol is normalized (comes inside of a 2x2 box in the range of [-1, 1])
        output = np.zeros((self.resolution, self.resolution), dtype=np.uint8)

        thickness = int(round(self.prc_width * self.resolution))
        unit_length = ((1.0 - self.prc_width) * self.resolution) / 2.0
        center_x = self.resolution / 2
        center_y = center_x

        for trace in symbol.traces:
            assert isinstance(trace, MathSymbolTrace)

            if len(trace.points) < 2:
                # single dot ...
                val_x, val_y = trace.points[0]
                p1_x = int(round(center_x + val_x * unit_length))
                p1_y = int(round(center_y + val_y * unit_length))

                # render as a single circle (dot)...
                cv2.circle(output, (p1_x, p1_y), thickness, (255,), thickness=-1)

            else:
                # scale points and compute trace length in scaled space ...
                total_length = 0.0
                scaled_points = []
                for idx, (val_x, val_y) in enumerate(trace.points):
                    curr_x = int(round(center_x + val_x * unit_length))
                    curr_y = int(round(center_y + val_y * unit_length))

                    if idx > 0:
                        _, prev_x, prev_y = scaled_points[-1]
                        total_length += math.sqrt(math.pow(curr_x - prev_x, 2) + math.pow(curr_y - prev_y, 2))

                    scaled_points.append((total_length, curr_x, curr_y))

                for idx in range(1, len(scaled_points)):
                    prev_l, prev_x, prev_y = scaled_points[idx - 1]
                    curr_l, curr_x, curr_y = scaled_points[idx]

                    mid_l = (prev_l + curr_l) / 2.0
                    segment_dist = math.sqrt(math.pow(curr_x - prev_x, 2) + math.pow(curr_y - prev_y, 2))

                    if total_length > 0.0:
                        prc_mid_l = mid_l / total_length
                    else:
                        prc_mid_l = 0.5

                    if self.decay_length <= prc_mid_l <= 1.0 - self.decay_length:
                        # normal part of the stroke
                        cv2.line(output, (prev_x, prev_y), (curr_x, curr_y), (255,), thickness)
                    else:
                        # with decay ...
                        if segment_dist < self.max_scaled_decay:
                            # segment is short enough, do in one trace ...
                            if prc_mid_l < self.decay_length:
                                # beginning ....
                                part_thickness = int(
                                    (prc_mid_l / self.decay_length) * thickness * self.decay_width + thickness * (
                                                1.0 - self.decay_width))
                            else:
                                # end ...
                                part_thickness = int(
                                    ((1.0 - prc_mid_l) / self.decay_length) * thickness * self.decay_width + thickness * (
                                                1.0 - self.decay_width))

                            cv2.line(output, (prev_x, prev_y), (curr_x, curr_y), (255,), part_thickness)
                        else:
                            # segment is long, do in parts ...
                            parts = int(math.ceil(segment_dist / self.max_scaled_decay))

                            for idx in range(parts):
                                pre_prc = (idx / parts)
                                next_prc = (idx + 1) / parts

                                p1_x = int(round(prev_x * (1.0 - pre_prc) + curr_x * pre_prc))
                                p1_y = int(round(prev_y * (1.0 - pre_prc) + curr_y * pre_prc))

                                p2_x = int(round(prev_x * (1.0 - next_prc) + curr_x * next_prc))
                                p2_y = int(round(prev_y * (1.0 - next_prc) + curr_y * next_prc))

                                w = (idx + 0.5) / parts
                                part_l = (prev_l * (1.0 - w) + curr_l * w) / total_length

                                if prc_mid_l < self.decay_length:
                                    # beginning ...
                                    part_thickness = int(
                                        (part_l / self.decay_length) * thickness * self.decay_width + thickness * (
                                                    1.0 - self.decay_width))
                                else:
                                    # end ...
                                    part_thickness = int(
                                        ((1.0 - part_l) / self.decay_length) * thickness * self.decay_width + thickness * (
                                                    1.0 - self.decay_width))

                                cv2.line(output, (p1_x, p1_y), (p2_x, p2_y), (255,), part_thickness)

        return output


