
import os
import csv
import shutil

import cv2


class QueryExporter:
    TemplatePath = "TangentV/resources/templates/queries"

    def __init__(self, binarization=None, source_validator=None):
        '''
        Initialize QueryExporter
        '''
        self.query = None
        self.query_set = None
        self.query_set_info = None
        self.tanv_query_set = None

        self.binarization = binarization
        self.source_validator = source_validator

        self.init_configs()

    def init_configs(self):
        '''
        This function initializes groups map, binarizer and other configs
        '''

        self.query = open(QueryExporter.TemplatePath + '/query.cfg', 'r', encoding='utf-8').read()
        self.query_set_info = open(QueryExporter.TemplatePath + '/query_set_info.cfg', 'r', encoding='utf-8').read()
        self.query_set = open(QueryExporter.TemplatePath + '/query_set.cfg', 'r', encoding='utf-8').read()
        self.tanv_query_set = open(QueryExporter.TemplatePath + '/query_file.cfg', 'r', encoding='utf-8').read()

    def load_query_index(self, query_index_filename, verbose=True):
        # get location of query index file
        index_dir, index_filename = os.path.split(query_index_filename)
        if index_dir == "":
            index_dir = "."

        # load query index file
        query_index = []
        with open(query_index_filename, 'r', encoding='utf8') as csvfile:
            file_reader = csv.reader(csvfile, delimiter=',')
            for row in file_reader:
                if len(row) < 3:
                    if verbose:
                        print("invalid query info: " + str(row))
                    continue

                query_id = row[0]
                query_path = index_dir + "/" + row[1]
                binary = row[2]

                if len(row) >= 4:
                    source = row[3]
                else:
                    source = "default"

                query_index.append({
                    "id": query_id,
                    "path": query_path,
                    "binary": binary,
                    "source": source,
                })

        return query_index

    def export(self, query_set_id, query_index, query_type, query_color, named_classifier, group_info, output_dir,
               verbose=True, copy_images=True):

        '''
        This function writes a access math query file to [output Directory]/[QuerySetID].xml
        '''
        if verbose:
            print('Started query export process')
        # create the output file
        os.makedirs(output_dir, exist_ok=True)

        # create dir for query set id where all images will be copied
        img_dir = output_dir + "/" + query_set_id
        if copy_images:
            os.makedirs(img_dir, exist_ok=True)

        query_groups = self.generate_groups(group_info)

        query_set = []

        for query_row in query_index:
            if verbose:
                print('Processing ' + str(query_row))

            # Validate the source
            if self.source_validator is not None:
                is_valid_source = self.source_validator.validate_source(query_row["source"])
                if verbose:
                    print(is_valid_source)
            else:
                is_valid_source = True

            if query_row["source"] is None or not is_valid_source:
                if verbose:
                    print('Source is not given or not valid. Using default source')
                source = 'default'

            # if image should be copied ...
            if copy_images:
                _, ext = os.path.splitext(query_row["path"])
                # copy output image ..
                output_img = img_dir + "/" + query_row["id"] + ext

                if query_row["binary"] == '0' and self.binarization is not None:
                    # if image is not binary, binarize it
                    tempo_image = cv2.imread(query_row["path"])
                    binary_img = 255 - self.binarization(tempo_image)
                    # save binary directly to output
                    cv2.imwrite(output_img, binary_img)
                else:
                    # copy source image to img_dir
                    shutil.copy(query_row["path"], output_img)

                query_path = query_set_id + "/" + query_row["id"] + ext
            else:
                # use source ...
                query_path = query_row["path"]

            if "metadata" in query_row:
                metadata = query_row["metadata"]
            else:
                metadata = None

            # create a query
            query = self.create_query(query_row["id"], query_path, query_groups, query_color, query_type,
                                      query_row["source"], metadata)
            query_set.append(query)

        query_set_str = self.create_query_set(query_set)
        query_set_info = self.create_query_set_info(query_set_id, named_classifier)

        tanv_query = self.create_tanv_query_set(query_set_info, query_set_str)

        # save to file
        out_filename = output_dir + "/" + query_set_id + ".xml"
        with open(out_filename, 'w', encoding='utf-8') as query_file:
            query_file.write(tanv_query)

        if verbose:
            print('Processing done.')

    def create_tanv_query_set(self, query_set_info, query_set):
        '''
        This function creates tangent-v query set xml file
        '''

        xml_query_set = self.tanv_query_set
        xml_query_set = xml_query_set.replace('__QUERY_SET_INFO__', query_set_info)
        xml_query_set = xml_query_set.replace('__QUERY_SET__', query_set)

        return xml_query_set

    def create_query_set(self, all_queries):
        '''
        This function creates a query set
        '''
        query_set_str = self.query_set
        query_set_str = query_set_str.replace('__QUERIES__', '\n'.join(all_queries))

        return query_set_str

    def create_query(self, query_id, query_path, query_groups, query_color, query_type, source, metadata=None):
        '''
        This function creates a query of the form
        <Query>
            <Id>Query-01</Id>
            <Type>Single-Image</Type>
            <Groups>
                    <Mode>[Attribute|Sequence]</Mode>
                    <GroupBy>[Attribute]</GroupBy>
                    [... <GroupBy>[Attribute]</GroupBy> ...]
                    [<SequenceBy>frame_idx</SequenceBy>]
            </Groups>
            <Images>
                    <Image>
                            <File>[QUERY PATH]</File>
                            <Color>[Query Color]</Color>
                            <Type>[Query type]</Type>
                            <Source>...</Source>
                            <MetaData>....</MetaData>
                    </Image>
            </Images>
        </Query>
        '''
        query = self.query
        query = query.replace('__QUERY_ID__', query_id)
        query = query.replace('__QUERY_PATH__', query_path)
        query = query.replace('__QUERY_COLOR__', query_color)
        query = query.replace('__QUERY_TYPE__', query_type)
        query = query.replace('__QUERY_SOURCE__', source)

        if metadata is None:
            query = query.replace('__QUERY_METADATA__', "")
        else:
            query = query.replace('__QUERY_METADATA__', metadata)

        # get group based on target domain
        # group_info
        query = query.replace('__GROUPS__', query_groups)

        return query

    def generate_labeled_boxes_metadata(self, file_data):
        ident = "                "
        symbol_template = '<Symbol bbox="{0:s} {1:s} {2:s} {3:s}" wildcard="{4:d}">{5:s}</Symbol>\n'
        label_template = '<label conf="{0:.2f}">{1:s}</label>'
        metadata = ""
        for x, y, w, h, label_dict, is_wildcard in file_data:
            label_str = ""
            for label in label_dict:
                label_str += label_template.format(label_dict[label], label)

            # not using format on floats here to avoid assuming/limiting floating point precision
            metadata += ident + symbol_template.format(str(x), str(y), str(w), str(h), 1 if is_wildcard else 0, label_str)

        return metadata

    def generate_groups(self, group_info):
        # Grouping mode
        ident = "                "
        xml = ident + "<Mode>{0:s}</Mode>\n".format(group_info["mode"])

        # Group attributes
        if "group_by" in group_info:
            for attribute in group_info["group_by"]:
                xml += ident + "<GroupBy>{0:s}</GroupBy>\n".format(attribute)

        # Sequence attributes
        if "sequence_by" in group_info:
            for attribute in group_info["sequence_by"]:
                xml += ident + "<SequenceBy>{0:s}</SequenceBy>\n".format(attribute)

        return xml

    def create_query_set_info(self, query_set_id, named_classifier):
        '''
        This function creates a query set info of the form

        <QuerySetInfo>
            <Id>__QUERY_SET_ID__</Id>
            <Name>__QUERY_SET_NAME__</Name>
            <Description>__SOURCE__ Image test query set containing single image queries only (Type 0 = Single-Image)</Description>
            <NamedClassifiers>__NAMED_CLASSIFIER__</NamedClassifiers>
        </QuerySetInfo>
        '''

        query_set_info = self.query_set_info

        query_set_info = query_set_info.replace('__QUERY_SET_ID__', query_set_id)
        query_set_info = query_set_info.replace('__NAMED_CLASSIFIER__', named_classifier)

        return query_set_info

