
import os
import pickle
import sys
import time

import cv2

from AM_CommonTools.util.time_helper import TimeHelper
from AccessMath.data.meta_data_DB import MetaDataDB
from AccessMath.indexing.lecture_class_mapping import LectureClassMapping

# TODO: these parameters could/should  come from DB configuration ...
from AccessMath.indexing.index_parameters import IndexParameters

from TangentV.util.latex_helper import LaTeX_Helper
from recognizer.configuration.configuration import Configuration

def main():
    if len(sys.argv) < 3:
        print("Usage:")
        print("")
        print(" python {0:s} database output_dir".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tdatabase:\tAccessMath Database XML file")
        print("\toutput_dir:\tOutput Directory where rendered traces will be stored ")
        return

    db_filename = sys.argv[1]
    output_directory = sys.argv[2]

    start_time = time.time()

    start_task = time.time()

    # Load database configuration ...
    try:
        am_database = MetaDataDB.from_file(db_filename)
    except:
        print("Invalid AccessMath database")
        return

    rec_config_filename = am_database.indexing.recognizer_config_filename

    try:
        rec_config = Configuration.from_file(rec_config_filename)
    except:
        print("Invalid Symbol Recognizer configuration")
        return

    # ... existing Full Mapping
    mapping_filename = rec_config.data["CLASSIFIER_DIR"] + "/per_lecture_classes.dat"
    print(mapping_filename)
    if os.path.isfile(mapping_filename):
        with open(mapping_filename, "rb") as in_file:
            full_mapping = pickle.load(in_file)
            assert isinstance(full_mapping, LectureClassMapping)
    else:
        print("Lecture-Mappings not found, must be created first")
        return

    end_task = time.time()
    loading_time = end_task - start_task

    start_task = time.time()

    # Extract Unique Classes
    # re-compute unique ...
    unique_labels = full_mapping.unique_classes
    print("Unique after mapping by shape: " + str(len(unique_labels)))
    print(unique_labels)

    # Map CROHME classes to Latex strings
    latex_strings = {}
    for label in unique_labels:
        if label in full_mapping.mapping_rec_to_latex:
            latex_strings[label] = full_mapping.mapping_rec_to_latex[label]
        else:
            latex_strings[label] = label

    # Generate images for each class
    ordered_labels = sorted(list(latex_strings.keys()))
    print("\nRendering Classes... ")

    images_per_label = LaTeX_Helper.render_latex_labels(ordered_labels, latex_strings,
                                                        IndexParameters.LTXRec_VerticalSizes,
                                                        IndexParameters.LTXRec_HorizontalSizes)

    end_task = time.time()
    initial_rendering_time = end_task - start_task

    start_task = time.time()

    images_per_label = LaTeX_Helper.expand_image_sets_per_label(images_per_label, IndexParameters.LTXRec_MinClassSample,
                                                                IndexParameters.LTXRec_MaxSizeDist)

    end_task = time.time()
    distortion_rendering_time = end_task - start_task

    # Save all images ...
    start_task = time.time()

    # create output directories
    # ... parent ...
    if not os.path.isdir(output_directory):
        os.mkdir(output_directory)

    # ... train sub ...
    current_dir = output_directory + "/training"
    if not os.path.isdir(current_dir):
        os.mkdir(current_dir)

    # ... per label ...
    mapping_lines = []
    for label_idx, label in enumerate(ordered_labels):
        # ... create directory (if doesn't exists) ...
        label_dir = current_dir + "/" + str(label_idx)
        if not os.path.isdir(label_dir):
            os.mkdir(label_dir)

        # add to current mapping ...
        mapping_lines.append(str(label_idx) + "\t" + label + "\n")

        # save images
        for img_idx, image in enumerate(images_per_label[label]):
            output_filename = "{0:s}/{1:d}/{2:d}.png".format(current_dir, label_idx, img_idx)
            cv2.imwrite(output_filename, image)

    # store class mapping/index ...
    mapping_filename = output_directory + "/class_mapping.txt"
    with open(mapping_filename, "w", encoding="utf8") as out_file:
        out_file.writelines(mapping_lines)

    end_task = time.time()
    saving_time = end_task - start_task

    end_time = time.time()
    print("Total Loading Time: " + TimeHelper.secondsToStr(loading_time))
    print("Total Initial Rendering Time: " + TimeHelper.secondsToStr(initial_rendering_time))
    print("Total Distortion Rendering Time: " + TimeHelper.secondsToStr(distortion_rendering_time))
    print("Total Saving Time: " + TimeHelper.secondsToStr(saving_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))

if __name__ == "__main__":
    main()