
import os
import pickle
import sys
import time

import numpy as np

from AM_CommonTools.util.time_helper import TimeHelper
from AccessMath.data.meta_data_DB import MetaDataDB
from AccessMath.indexing.lecture_class_mapping import LectureClassMapping

from TangentV.indexing.recognition_helper import RecognitionHelper

from recognizer.data.feature_dataset import FeatureDataset
from recognizer.evaluation.eval_ops import EvalOps
from recognizer.training.trainer import ClassifierTrainer
from recognizer.configuration.configuration import Configuration


def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} database".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tdatabase:\tAccessMath Database XML file")
        return

    db_filename = sys.argv[1]

    start_time = time.time()

    start_task = time.time()
    # Load database configuration ...
    try:
        am_database = MetaDataDB.from_file(db_filename)
    except:
        print("Invalid AccessMath database")
        return

    rec_config_filename = am_database.indexing.recognizer_config_filename
    named_classifiers_filename = am_database.indexing.named_latex_classifiers_filename

    try:
        rec_config = Configuration.from_file(rec_config_filename)
    except:
        print("Invalid Symbol Recognizer configuration")
        return

    # ... existing Full Mapping
    mapping_filename = rec_config.data["CLASSIFIER_DIR"] + "/per_lecture_classes.dat"
    print(mapping_filename)
    if os.path.isfile(mapping_filename):
        with open(mapping_filename, "rb") as in_file:
            full_mapping = pickle.load(in_file)
            assert isinstance(full_mapping, LectureClassMapping)
    else:
        print("Lecture-Mappings not found, must be created first")
        return

    training_filename = rec_config.data["DATASET_DIR"] + "/ds_LATEX_train.dat"
    dataset = FeatureDataset.load_from_file(training_filename)
    # copy raw labels ...
    training_labels = dataset.labels
    RecognitionHelper.set_dataset_mapping(dataset, None, None)

    end_task = time.time()
    loading_time = end_task - start_task

    start_task = time.time()

    # Train Math Symbol Classifier
    print("-> Training General Classifier ....")
    classifier = ClassifierTrainer.train_classifier(dataset, rec_config)

    end_task = time.time()
    training_time = end_task - start_task

    start_task = time.time()

    # Evaluate on Latex Training Data
    train_acc, train_avg, train_std, _ = EvalOps.classifier_per_class_accuracy(classifier, dataset)

    print("-> General Classifier Training Results")
    print("\tGlb. ACC\tPC. AVG\tPC. STD")
    print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(train_acc, train_avg, train_std))

    end_task = time.time()
    evaluation_time = end_task - start_task

    # Save to file
    print("-> Saving General classifier")
    out_filename = rec_config.data["CLASSIFIER_DIR"] + "/" + rec_config.data["CLASSIFIER_NAME"] + "_LATEX.dat"
    named_classifiers = [("default", rec_config_filename, out_filename)]
    out_file = open(out_filename, 'wb')
    pickle.dump(classifier, out_file, pickle.HIGHEST_PROTOCOL)
    out_file.close()

    # now, repeat the process per lecture ...
    preparation_time = 0.0
    all_train_acc, all_train_avg, all_train_std = [], [], []
    print("\nTraining a classifier per lecture ...")
    for idx in full_mapping.lecture_mapping:
        lecture_title = full_mapping.lecture_mapping[idx].title
        print("\nProcessing: " + lecture_title, flush=True)

        task_start = time.time()
        # first, get the set of unique symbols from CROHME that appear on this lecture (according to transcripts)
        lecture_symbols = full_mapping.classes_per_lecture[lecture_title]

        print("> Total Lecture Classes: " + str(len(lecture_symbols)))

        # split datasets based on the symbols found in the lecture ...
        # first, restore raw labels before splitting ...
        dataset.labels = training_labels
        lecture_dataset, extra_dataset = RecognitionHelper.split_training_set_by_classes(dataset, lecture_symbols)

        print(lecture_dataset.data.shape)

        # print some stats....
        n_training_samples, n_training_features = lecture_dataset.data.shape
        n_extra_samples = extra_dataset.data.shape[0]
        prc_training = n_training_samples / dataset.data.shape[0]
        prc_extra = n_extra_samples / dataset.data.shape[0]

        print("-> Total features: {0:d}".format(n_training_features))

        print("-> Training Symbols")
        print("---> from Lecture Classes: {0:d} ({1:.2f}%)".format(n_training_samples, prc_training * 100.0))
        print("---> from Other Classes: {0:d} ({1:.2f}%)".format(n_extra_samples, prc_extra * 100.0))

        RecognitionHelper.set_dataset_mapping(lecture_dataset)
        task_end = time.time()
        preparation_time += task_end - task_start

        # train a classifier using only symbols that appear on math expressions from current lecture...
        task_start = time.time()
        print("-> Training ...", flush=True)
        
        # train classifier ...
        classifier = ClassifierTrainer.train_classifier(lecture_dataset, rec_config)
        task_end = time.time()
        training_time += task_end - task_start
        
        task_start = time.time()
        # evaluate on lecture training sets
        train_acc, train_avg, train_std, _ = EvalOps.classifier_per_class_accuracy(classifier, lecture_dataset)

        print("-> Training Results")
        print("\tGlb. ACC\tPC. AVG\tPC. STD")
        print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(train_acc, train_avg, train_std))

        # add for global averages ...
        all_train_acc.append(train_acc)
        all_train_avg.append(train_avg)
        all_train_std.append(train_std)

        task_end = time.time()
        evaluation_time += task_end - task_start

        print("-> Saving classifier")
        out_filename = rec_config.data["CLASSIFIER_DIR"] + "/" + rec_config.data["CLASSIFIER_NAME"] + "_LATEX_" + lecture_title + ".dat"
        named_classifiers.append((lecture_title, rec_config_filename, out_filename))

        out_file = open(out_filename, 'wb')
        pickle.dump(classifier, out_file, pickle.HIGHEST_PROTOCOL)
        out_file.close()

    print("\t\tGlobal Averages")
    print("Training Results")
    print("\tGlb. ACC\tPC. AVG\tPC. STD")
    print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(np.array(all_train_acc).mean(), np.array(all_train_avg).mean(),
                                               np.array(all_train_std).mean()))

    print("\n\nSaving Named Classifiers Index to: " + named_classifiers_filename)
    with open(named_classifiers_filename, "w", encoding="utf8") as out_file:
        for rec_name, conf_filename, rec_path in named_classifiers:
            out_file.write("{0:s}\t{1:s}\t{2:s}\n".format(rec_name, conf_filename, rec_path))

    end_time = time.time()
    print("Total Loading Time: " + TimeHelper.secondsToStr(loading_time))
    print("Total preparation Time: " + TimeHelper.secondsToStr(preparation_time))
    print("Total training Time: " + TimeHelper.secondsToStr(training_time))
    print("Total evaluation Time: " + TimeHelper.secondsToStr(evaluation_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))


if __name__ == "__main__":
    main()
