import os
import pickle
import sys
import time

import cv2

from AM_CommonTools.util.time_helper import TimeHelper
from AccessMath.annotation.keyframe_annotation import KeyFrameAnnotation
from AccessMath.data.meta_data_DB import MetaDataDB

from AM_CommonTools.configuration.configuration import Configuration as ConfigTanv
from recognizer.configuration.configuration import Configuration as ConfigRec


def load_keyframes(database):
    all_keyframes = {}

    for lecture in database.lectures:
        summary_prefix = database.output_summaries + "/" + database.name + "_" + lecture.title.lower()

        summ_filename = summary_prefix + "/segments.xml"
        summ_image_prefix = summary_prefix + "/keyframes/"

        # ideal summary ...
        if os.path.exists(summ_filename):
            print("Summary found: " + lecture.title.lower())

            summ_keyframes = KeyFrameAnnotation.LoadExportedKeyframes(summ_filename, summ_image_prefix, False, False, True)
            all_keyframes[lecture.title] = {"lecture": lecture, "keyframes": summ_keyframes}
        else:
            print("Summary Not found: " + lecture.title.lower())

    return all_keyframes


def main():
    if len(sys.argv) < 2:
        print("Usage:")
        print("")
        print(" python {0:s} database conf_file".format(sys.argv[0]))
        print("")
        print("Where")
        print("\tDatabase:\tPath to AccessMath database file (in XML format)")
        return

    database_filename = sys.argv[1]

    start_time = time.time()

    # load database
    try:
        database = MetaDataDB.from_file(database_filename)
    except:
        print("Invalid database file")
        return

    rec_config_filename = database.indexing.recognizer_config_filename
    tanv_config_filename = database.indexing.image_mir_config_filename
    named_classifiers_filename = database.indexing.named_hw_classifiers_filename
    allow_default_rec = database.indexing.allow_default_recognizer

    # load recognizer config ...
    try:
        rec_config = ConfigRec.from_file(rec_config_filename)
    except:
        print("Invalid Symbol Recognizer configuration")
        return

    base_classifier_name = rec_config.data["CLASSIFIER_DIR"] + "/" + rec_config.data["CLASSIFIER_NAME"]
    default_classifier_filename = base_classifier_name + ".dat"

    # load tangent-V config ...
    try:
        tanv_config = ConfigTanv.from_file(tanv_config_filename)
    except:
        print("Invalid Tangent-V configuration")
        return

    # load key-frames from summaries ...
    all_keyframes = load_keyframes(database)

    task_end = time.time()
    total_loading_time = task_end - start_time

    task_start = time.time()

    # create the export paths ....
    # ... main ....
    main_export_dir = database.indexing.hw_image_path
    main_index_filename = main_export_dir + "/image_index.dat"
    if not os.path.isdir(main_export_dir):
        # create it ...
        os.mkdir(main_export_dir)

    # .... images .....
    export_image_dir = main_export_dir + "/images"
    export_image_index_filename = export_image_dir + "/index.txt"
    if not os.path.isdir(export_image_dir):
        os.mkdir(export_image_dir)
    # .... searches  .....
    export_search_dir = main_export_dir + "/searches"
    if not os.path.isdir(export_search_dir):
        os.mkdir(export_search_dir)
    # .... temporal  .....
    export_temporal_dir = main_export_dir + "/temporal"
    if not os.path.isdir(export_temporal_dir):
        os.mkdir(export_temporal_dir)

    default_classifier_info = (rec_config_filename, default_classifier_filename)

    # export custom configuration
    # ... set new config parameters ....
    tanv_config.set("BINARY_IMAGES_DIRECTORY", export_image_dir)
    tanv_config.set("BINARY_IMAGES_METADATA", export_image_index_filename)
    tanv_config.set("INDEX_FILENAME", main_index_filename)
    tanv_config.set("SEARCH_OUTPUT_DIRECTORY", export_search_dir)
    tanv_config.set("CLASSIFIERS_CONFIG_DIRECTORY", ".")
    tanv_config.set("CLASSIFIERS_TRAINED_DIRECTORY", ".")
    tanv_config.set("TEMPORARY_CLASSIFIED_SYMBOLS", export_temporal_dir + "/classified_symbols_")
    tanv_config.set("SYMBOL_CLASSIFIER_DEFAULT", default_classifier_info)
    tanv_config.set("SYMBOL_CLASSIFIER_BY_NAME", named_classifiers_filename)
    tanv_config.set("HTML_LINK_GENERATOR", "AccessMath.indexing.tangentv_helper")
    tanv_config.set("HTML_LINK_PARAMS", {"database": database_filename})
    tanv_config.set("HTML_MATCH_CUT_PADDING", (100, 100, 100, 100))
    tanv_config.set("METADATA_INDEX_BY", ["lecture"])
    # ... save ....
    tanv_config.save(database.indexing.hw_tanv_config_path)

    # export key-frames per lecture
    image_index_header = "instance_id\timg_location\tclassifier\tlecture\tframe_idx\n"
    image_index_lines = [image_index_header]
    for lecture_title in all_keyframes:
        print("\n\nProcessing: " + lecture_title)
        lecture = all_keyframes[lecture_title]["lecture"]
        keyframes = all_keyframes[lecture_title]["keyframes"]

        # Check if classifier exists ....
        classifier_filename = base_classifier_name + "_" + lecture_title + ".dat"

        if os.path.exists(classifier_filename):
            # valid classifier, use name
            classifier_name = lecture_title
        else:
            if allow_default_rec:
                classifier_name = "default"
            else:
                print("WARNING: classifier not found! Skipping lecture!!!")
                print("")
                continue

        # dir for lecture images ...
        lecture_dir = export_image_dir + "/" + lecture_title
        if not os.path.isdir(lecture_dir):
            os.mkdir(lecture_dir)

        # .... images ...
        for kf_idx, kf in enumerate(keyframes):
            image_suffix_name = str(kf_idx) + ".png"
            image_filename = lecture_dir + "/" + image_suffix_name
            cv2.imwrite(image_filename, kf.binary_image)

            instance_id = lecture_title + ":" + str(kf.idx)
            image_location = lecture_title + "/" + image_suffix_name
            image_line = "\t".join([instance_id, image_location, classifier_name, lecture_title, str(kf.idx)]) + "\n"
            image_index_lines.append(image_line)

    # Save image index
    with open(export_image_index_filename, "w", encoding="utf8") as out_file:
        out_file.writelines(image_index_lines)

    task_end = time.time()
    total_export_time = task_end - task_start

    end_time = time.time()

    print("\n")
    print("Loading Time: " + TimeHelper.secondsToStr(total_loading_time))
    print("Export Time: " + TimeHelper.secondsToStr(total_export_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))
    print("Finished!")

if __name__ == '__main__':
    main()