
import math
import numpy as np

class HistogramFeatureExtractor:
    @staticmethod
    def trace2DLengthHistogram(trace, rows, cols, min_x, min_y, max_x, max_y):
        # create the bins...
        distribution = np.zeros((rows, cols), np.float64)

        bin_size_x = (max_x - min_x) / (cols - 1)
        bin_size_y = (max_y - min_y) / (rows - 1)

        for i in range(len(trace.points) - 1):
            p0_x, p0_y = trace.points[i]
            p1_x, p1_y = trace.points[i + 1]

            mid_x = (p0_x + p1_x) / 2.0
            mid_y = (p0_y + p1_y) / 2.0

            if min_x <= mid_x <= max_x and min_y <= mid_y <= max_y:
                length = math.sqrt((p1_x - p0_x) ** 2 + (p1_y - p0_y) ** 2)

                h_div = (mid_x - min_x) / bin_size_x
                h_bin0 = int(math.floor(h_div))
                if h_bin0 == cols - 1:
                    h_bin0 = cols - 2
                    h_w1 = 1.0
                else:
                    h_w1 = h_div - h_bin0
                h_bin1 = h_bin0 + 1

                v_div = (mid_y - min_y) / bin_size_y
                v_bin0 = int(math.floor(v_div))
                if v_bin0 == rows - 1:
                    v_bin0 = rows - 2
                    v_w1 = 1.0
                else:
                    v_w1 = v_div - v_bin0
                v_bin1 = v_bin0 + 1

                distribution[v_bin0, h_bin0] += length * (1.0 - h_w1) * (1.0 - v_w1)
                distribution[v_bin0, h_bin1] += length * h_w1 * (1.0 - v_w1)
                distribution[v_bin1, h_bin0] += length * (1.0 - h_w1) * v_w1
                distribution[v_bin1, h_bin1] += length * h_w1 * v_w1

        return distribution

    @staticmethod
    def trace2DPointHistogram(trace, rows, cols, min_x, min_y, max_x, max_y):
        # create the bins...
        distribution = np.zeros((rows, cols), np.float64)

        bin_size_x = (max_x - min_x) / (cols - 1)
        bin_size_y = (max_y - min_y) / (rows - 1)

        for x, y in trace.points:
            if min_x <= x <= max_x and min_y <= y <= max_y:
                h_div = (x - min_x) / bin_size_x
                h_bin0 = int(math.floor(h_div))
                if h_bin0 == cols - 1:
                    h_bin0 = cols - 2
                    h_w1 = 1.0
                else:
                    h_w1 = h_div - h_bin0
                h_bin1 = h_bin0 + 1

                v_div = (y - min_y) / bin_size_y
                v_bin0 = int(math.floor(v_div))
                if v_bin0 == rows - 1:
                    v_bin0 = rows - 2
                    v_w1 = 1.0
                else:
                    v_w1 = v_div - v_bin0
                v_bin1 = v_bin0 + 1

                distribution[v_bin0, h_bin0] += (1.0 - h_w1) * (1.0 - v_w1)
                distribution[v_bin0, h_bin1] += h_w1 * (1.0 - v_w1)
                distribution[v_bin1, h_bin0] += (1.0 - h_w1) * v_w1
                distribution[v_bin1, h_bin1] += h_w1 * v_w1

        return distribution

    @staticmethod
    def computeSymbol2DHistogram(symbol, grids):
        features = []
        for n, m, w_prc, h_prc, off_x, off_y, use_lengths in grids:
            min_x = -1.0 + off_x * 2.0
            min_y = -1.0 + off_y * 2.0
            max_x = min_x + 2.0 * w_prc
            max_y = min_y + 2.0 * h_prc

            # ...for each trace...
            bidimensional_hist = np.zeros((n, m), dtype=np.float64)
            for trace in symbol.traces:
                if use_lengths:
                    hist = HistogramFeatureExtractor.trace2DLengthHistogram(trace, n, m, min_x, min_y, max_x, max_y)
                else:
                    hist = HistogramFeatureExtractor.trace2DPointHistogram(trace, n, m, min_x, min_y, max_x, max_y)

                bidimensional_hist += hist


            # then, normalize!
            normalization_factor = bidimensional_hist.sum()
            if normalization_factor > 0.0:
                bidimensional_hist /= normalization_factor

            # add to feature vector (as list of continuous attributes)
            features += bidimensional_hist.reshape((n * m,)).tolist()

        return features


    @staticmethod
    def getTraceOrientationHistogram(trace, rows, cols, min_x, min_y, max_x, max_y, orientation_bins):
        distribution = np.zeros((rows, cols, orientation_bins), np.float64)

        distances = []
        mid_points = []
        orientations = []
        for i in range(len(trace.points) - 1):
            p0_x, p0_y = trace.points[i]
            p1_x, p1_y = trace.points[i + 1]
            mid_x = (p0_x + p1_x) / 2
            mid_y = (p0_y + p1_y) / 2

            # only if the middle point of the segment is inside of the current grid
            if min_x <= mid_x <= max_x and min_y <= mid_y <= max_y:
                mid_points.append((mid_x, mid_y))

                # Length
                length = math.sqrt((p1_x - p0_x) ** 2 + (p1_y - p0_y) ** 2)
                distances.append(length)

                # Orientation
                currentAngle = math.atan2(p1_y - p0_y, p1_x - p0_x)
                orientations.append(currentAngle)

        # Now calculate the angles...
        ang_bin_size = math.pi / orientation_bins
        off = math.pi

        grid_w = max_x - min_x
        grid_h = max_y - min_y

        # use midpoint of the line to distribute values on the corresponding cells of the grid
        for i, (mid_x, mid_y) in enumerate(mid_points):
            # relative weight according to length in relation to total...
            wl = distances[i]

            if rows >= 2:
                #....for Y....
                val_y = ((mid_y - min_y) / grid_h) * (rows - 1)
                c0_y = int(val_y)
                if c0_y == rows - 1:
                    c0_y -= 1
                c1_y = c0_y + 1
                w1_y = val_y - c0_y
                w0_y = 1 - w1_y

            if cols >= 2:
                #....for X....
                val_x = ((mid_x - min_x) / grid_w) * (cols - 1)
                c0_x = int(val_x)
                if c0_x == cols - 1:
                    c0_x -= 1
                c1_x = c0_x + 1
                w1_x = val_x - c0_x
                w0_x = 1 - w1_x

            # angle of current segment of line...
            currentAngle = orientations[i]

            # angle is between -pi and pi, add offset and divide between angular bin size
            p = (currentAngle + off) / ang_bin_size
            # the base orientation
            fp = math.floor(p)

            # the weight of the second orientation
            wp1 = p - fp
            # select bin for current orientation
            p0 = int(fp % orientation_bins)
            # select bin for next orientation (circular)
            p1 = int((p0 + 1) % orientation_bins)

            # the values....
            g0 = wl * (1.0 - wp1)
            g1 = wl * wp1

            # add to corresponding bins...
            if rows >= 2 and cols >= 2:
                # 4 cells affected, 8 bins
                # (0,0)
                distribution[c0_y][c0_x][p0] += g0 * w0_x * w0_y
                distribution[c0_y][c0_x][p1] += g1 * w0_x * w0_y
                # (0,1)
                distribution[c0_y][c1_x][p0] += g0 * w1_x * w0_y
                distribution[c0_y][c1_x][p1] += g1 * w1_x * w0_y
                # (1,0)
                distribution[c1_y][c0_x][p0] += g0 * w0_x * w1_y
                distribution[c1_y][c0_x][p1] += g1 * w0_x * w1_y
                # (1,1)
                distribution[c1_y][c1_x][p0] += g0 * w1_x * w1_y
                distribution[c1_y][c1_x][p1] += g1 * w1_x * w1_y
            elif rows >= 2:
                # 2 cells affected, 4 bins
                # (0,0)
                distribution[c0_y][0][p0] += g0 * w0_y
                distribution[c0_y][0][p1] += g1 * w0_y
                # (1,0)
                distribution[c1_y][0][p0] += g0 * w1_y
                distribution[c1_y][0][p1] += g1 * w1_y
            elif cols >= 2:
                # 2 cells affected, 4 bins
                # (0,0)
                distribution[0][c0_x][p0] += g0 * w0_x
                distribution[0][c0_x][p1] += g1 * w0_x
                # (0,1)
                distribution[0][c1_x][p0] += g0 * w1_x
                distribution[0][c1_x][p1] += g1 * w1_x
            else:
                # 1 cell affected, 2 bins
                distribution[0][0][p0] += g0
                distribution[0][0][p1] += g1

        return distribution

    @staticmethod
    def computeSymbolOrientationHistogram(symbol, grids):
        features = []
        for n, m, w_prc, h_prc, off_x, off_y, orientation_bins in grids:
            min_x = -1.0 + off_x * 2.0
            min_y = -1.0 + off_y * 2.0
            max_x = min_x + 2.0 * w_prc
            max_y = min_y + 2.0 * h_prc

            # ...for each trace...
            bidimensional_hist = np.zeros((n, m, orientation_bins), np.float64)
            for trace in symbol.traces:
                hist = HistogramFeatureExtractor.getTraceOrientationHistogram(trace, n, m, min_x, min_y, max_x, max_y,
                                                                              orientation_bins)

                bidimensional_hist += hist

            # then, normalize!
            normalization_factor = bidimensional_hist.sum()
            if normalization_factor > 0.0:
                bidimensional_hist /= normalization_factor

            # add to feature vector (as list of continuous attributes)
            features += bidimensional_hist.reshape((n * m * orientation_bins,)).tolist()

        return features

