
import os
import sys
import time
import pickle
import codecs

from AM_CommonTools.util.time_helper import TimeHelper
from AM_CommonTools.configuration.configuration import Configuration

from TangentV.indexing.index_parameters import IndexParameters
from TangentV.indexing.graph_index import GraphIndex
from TangentV.retrieval.query_set import QuerySet


def main():
    if len(sys.argv) < 3:
        print("Usage:")
        print("")
        print(" python tanv_02_search.py tanv_config queries [metric] [query_set_id]")
        print("")
        print("Where")
        print("\ttanv_conf\tTangent-V Configuration File")
        print("\tqueries\t\tFile defining queries")
        print("")
        print("Optional")
        print("\tmetric\t\tOverride configuration similarity metric")
        print("\tid\t\tOverride override query set id")
        return

    tanv_config_filename = sys.argv[1]
    query_filename = sys.argv[2]

    start_time = time.time()

    # Loading ....
    task_start = time.time()

    try:
        tanv_config = Configuration.from_file(tanv_config_filename)
    except Exception as error:
        print("Invalid Tangent-v configuration")
        print(error)
        return

    if len(sys.argv) >= 4:
        # override configuration ...
        print("Override Similarity metric: {0:d} to {1:s}".format(tanv_config.get("SIMILARITY_METRIC"), sys.argv[3]))
        tanv_config.set("SIMILARITY_METRIC", sys.argv[3])

    try:
        IndexParameters.ReadFromConfiguration(tanv_config)
    except Exception as error:
        print("Invalid Tangent-v configuration")
        print(error)
        return

    # check for usage of external link generator ...
    if tanv_config.contains("HTML_LINK_GENERATOR"):
        # load external module ...
        print("... Loading HTML link generator ...")
        generator_module = __import__(tanv_config.get("HTML_LINK_GENERATOR"), fromlist=["get_link_generator"])
        link_generator = generator_module.get_link_generator()

        if tanv_config.contains("HTML_LINK_PARAMS"):
            link_params = tanv_config.get("HTML_LINK_PARAMS")
        else:
            link_params = {}

        # ... set external module ...
        link_generator.prepare(link_params)
    else:
        # no link generator class
        link_generator = None

    # Load Queries
    query_set = QuerySet.FromXML(query_filename, tanv_config)

    if len(sys.argv) >= 4:
        # override query set id
        print("Override Query set id from {0:s} to {1:s}".format(query_set.id, sys.argv[4]))
        query_set.id = sys.argv[4]

    print("Loaded Query Set: " + query_set.id)
    print("-> Total queries: " + str(len(query_set.queries)))

    # Load index
    index_filename = tanv_config.get("INDEX_FILENAME")
    print("Loading index: " + index_filename)
    image_index = GraphIndex.LoadFromFile(index_filename)
    assert isinstance(image_index, GraphIndex)

    print("=> Total Unique Images: {0:d}".format(image_index.total_images()))
    print("=> Total Source IDs: {0:d}".format(image_index.total_source_ids()))
    print("=> Total Unique Pairs: {0:d}".format(image_index.total_pairs()))
    print("=> Total Pair Instances: {0:d}".format(image_index.total_instances()))

    print("Optimizing index ...")
    image_index.optimize_for_search()

    task_end = time.time()
    total_loading_time = task_end - task_start

    # Query pre-processing
    task_start = time.time()
    query_set.preprocess(tanv_config)
    task_end = time.time()
    total_query_preprocessing_time = task_end - task_start

    # Execute Queries
    print("\n\nExecuting Queries....")
    task_start = time.time()
    query_set.execute(image_index)
    task_end = time.time()
    total_query_execution_time = task_end - task_start

    # save results
    task_start = time.time()

    output_HTML_K = tanv_config.get("SEARCH_OUTPUT_HTML_MAX")
    output_BBOX_K = tanv_config.get("SEARCH_OUTPUT_BBOX_MAX")
    output_TREC_K = tanv_config.get("SEARCH_OUTPUT_TREC_MAX")

    if tanv_config.contains("BINARY_IMAGES_FORMAT"):
        image_format = tanv_config.get("BINARY_IMAGES_FORMAT")
    else:
        # by default, assume binary png ...
        image_format = "png"

    query_set.save_TREC_results(output_TREC_K)
    query_set.save_BBOX_results(output_BBOX_K)

    if image_format == "png":
        query_set.save_HTML_results(output_HTML_K, link_generator)
    else:
        print("HTML output for : " + image_format + " might fail", flush=True)
        query_set.save_HTML_results(output_HTML_K, link_generator)

    query_set.save_stats()
    task_end = time.time()
    total_results_generation_time = task_end - task_start

    end_time = time.time()

    # Show query statistics
    print("Loading Time: " + TimeHelper.secondsToStr(total_loading_time))
    print("Query Pre-processing Time: " + TimeHelper.secondsToStr(total_query_preprocessing_time))
    print("Query Execution Time: " + TimeHelper.secondsToStr(total_query_execution_time))
    print("Query Results generation Time:" + TimeHelper.secondsToStr(total_results_generation_time))
    print("Total Elapsed Time: " + TimeHelper.secondsToStr(end_time - start_time))

    print("Finished!")


if __name__ == "__main__":
    if sys.stdout.encoding != 'utf8':
      sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
    if sys.stderr.encoding != 'utf8':
      sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')

    main()

